module Spec.Claim exposing
  ( Claim
  , Verdict(..)
  , satisfying
  , isTrue
  , isFalse
  , isEqual
  , isListWithLength
  , isListWhere
  , isListWhereItemAt
  , isSomething
  , isNothing
  )

{-| A claim is a function that maps a subject to a verdict.

@docs Claim, Verdict

# Basic Claims
@docs isEqual, isTrue, isFalse

# Claims about Lists
@docs isListWhere, isListWhereItemAt, isListWithLength

# Claims about Maybe types
@docs isSomething, isNothing

# Combining Claims
@docs satisfying

-}

import Spec.Report as Report exposing (Report)


{-| Represents a function from a subject to a verdict.
-}
type alias Claim a =
  a -> Verdict


{-| The result of applying a claim to a subject.

A claim about a subject will either be accepted or rejected.
-}
type Verdict
  = Accept
  | Reject Report


{-| Combine multiple claims into one.

If any of the claims is rejected, then the combined claim is rejected.
-}
satisfying : List (Claim a) -> Claim a
satisfying claims actual =
  List.foldl (\claim verdict ->
    case claim actual of
      Accept ->
        verdict
      Reject report ->
        case verdict of
          Accept ->
            Reject <| Report.batch
              [ Report.note "Expected all claims to be satisfied, but one or more were rejected"
              , report
              ]
          Reject existingReport ->
            Reject <| Report.batch
              [ existingReport
              , Report.note "and"
              , report
              ]
  ) Accept claims


{-| Claim that the subject is `True`
-}
isTrue : Claim Bool
isTrue =
  isEqual boolWriter True


{-| Claim that the subject is `False`
-}
isFalse : Claim Bool
isFalse =
  isEqual boolWriter False


boolWriter : Bool -> String
boolWriter b =
  if b == True then
    "True"
  else
    "False"


{-| Claim that the subject is equal to the provided value.

The first argument to this function converts a value of some type to a string, which
is used to produce readable messages if the claim is rejected. Instead of providing
this string-generating function each time, I suggest adding a helper function that just uses the
`Debug.toString` function like so:

    equals : a -> Claim a
    equals =
      Spec.Claim.isEqual Debug.toString

Then just use your `equals` function whenever you need to claim that a subject is equal to some value.
-}
isEqual : (a -> String) -> a -> Claim a
isEqual toString expected actual =
  if expected == actual then
    Accept
  else
    Reject <| Report.batch
      [ Report.fact "Expected" <| toString actual
      , Report.fact "to equal" <| toString expected
      ]


{-| Claim that the subject is a list with the given length.
-}
isListWithLength : Int -> Claim (List a)
isListWithLength expected actual =
  let
    actualLength = List.length actual
  in
    if actualLength == expected then
      Accept
    else
      Reject <| wrongLength expected actualLength


{-| Claim that the subject is a list where the following claims are satisfied:

- the subject has the same length as the provided list
- for each item in the subject, that item satisfies the corresponding claim in the provided list.

For example:

    [ 1, 2, 3 ]
      |> Spec.Claim.isListWhere
          [ Spec.Claim.isEqualTo Debug.toString 1
          , Spec.Claim.isEqualTo Debug.toString 27
          , Spec.Claim.isEqualTo Debug.toString 3
          ]

would result in a rejected claim, since 2 is not equal to 27.
-}
isListWhere : List (Claim a) -> Claim (List a)
isListWhere claims actual =
  if List.length claims == List.length actual then
    matchList 1 claims actual
  else
    Reject <| Report.batch
      [ Report.note "List failed to match"
      , wrongLength (List.length claims) (List.length actual)
      ]


wrongLength : Int -> Int -> Report
wrongLength expected actual =
  Report.batch
  [ Report.fact "Expected list to have length" <| String.fromInt expected
  , Report.fact "but it has length" <| String.fromInt actual
  ]


matchList : Int -> List (Claim a) -> Claim (List a)
matchList position claims actual =
  case ( claims, List.head actual ) of
    ( [], Nothing ) ->
      Accept
    ( next :: remaining, Just head ) ->
      case next head of
        Accept ->
          matchList (position + 1) remaining <| List.drop 1 actual
        Reject report ->
          Reject <| Report.batch
            [ Report.note <| "List failed to match at position " ++ String.fromInt position
            , report
            ]
    _ ->
      Reject <| Report.note "Something crazy happened"


{-| Claim that the subject is a list such that:

- there is an item at the given index
- that item satisfies the given claim
-}
isListWhereItemAt : Int -> Claim a -> Claim (List a)
isListWhereItemAt index claim actualList =
  case List.head <| List.drop index actualList of
    Just actual ->
      claim actual
        |> mapRejection (\report -> Report.batch
            [ Report.note <| "Item at index " ++ String.fromInt index ++ " did not satisfy claim:"
            , report
            ]
        )
    Nothing ->
      Reject <| Report.batch
        [ Report.fact "Expected item at index" <| String.fromInt index
        , Report.fact "but the list has length" <| String.fromInt <| List.length actualList
        ]


mapRejection : (Report -> Report) -> Verdict -> Verdict
mapRejection mapper verdict =
  case verdict of
    Accept ->
      Accept
    Reject report ->
      Reject <| mapper report


{-| Claim that the subject is the `Just` case of the `Maybe` type.
-}
isSomething : Claim (Maybe a)
isSomething actual =
  case actual of
    Just _ ->
      Accept
    Nothing ->
      Reject <| Report.batch
        [ Report.fact "Expected" "something"
        , Report.fact "but found" "nothing"
        ]


{-| Claim that the subject is the `Nothing` case of the `Maybe` type.
-}
isNothing : Claim (Maybe a)
isNothing actual =
  case actual of
    Just _ ->
      Reject <| Report.batch
        [ Report.fact "Expected" "nothing"
        , Report.fact "but found" "something"
        ]
    Nothing ->
      Accept